/*
	Sample C WIN32 application using clipboard
	Neuron Computing Inc.
*/

#include <windows.h>

#include "Clipboard.h"

TCHAR s_szSampleText[] = 
	TEXT("Clipboard sample program demonstrates cut, copy and paste ")
	TEXT("clipboard operations and the appropriate WIN32 API calls.\r\n")
	TEXT("The application also demonstrates the use of ")
	TEXT("MENU, ACCELERATORS and DIALOG BOX resources.\n\r")
	TEXT("Its entry point WinMain creates a WNDCLASS, loads the ")
	TEXT("accelerators, creates a main window (WS_OVERLAPPEDWINDOW) ")
	TEXT("and enters event loop.\r\n")
	TEXT("\r\n")
	TEXT("The application uses DrawText GDI32 call to paint text into its ")
	TEXT("client area.\r\n")
	TEXT("\r\n")
	TEXT("The applications demonstrates a simple About dialog box with ")
	TEXT("labels and a push button.\r\n")
	TEXT("\r\n")
	TEXT("The application uses GlobalAlloc, GlobalLock, GlobalUnlock and ")
	TEXT("GlobalFree WIN32 API calls instead of malloc and free C run ")
	TEXT("time library functions.\r\n")
	TEXT("\r\n")
	TEXT("The application uses UNICODE strings if compiled with UNICODE ")
	TEXT("defined.\r\n")
	TEXT("\r\n")
	TEXT("Use Edit->Copy or Edit->Cut menu command to move this text to ")
	TEXT("the clipboard.\r\n")
	TEXT("Use Edit->Paste menu command to replace this text ")
	TEXT("with data posted to the clibboard by another application.");

/* Global variables */
HINSTANCE g_hInstance = NULL;
HWND g_hWnd = NULL;
HACCEL g_hAccTable = NULL;
HANDLE g_hText = NULL;

/*
	Out of memory error
*/
void Error_OutOfMemory()
{
	HWND hWndFocus = GetFocus();
	MessageBox(hWndFocus, TEXT("Out of Memory"),
				NULL, MB_ICONHAND | MB_SYSTEMMODAL);
    return;
}

/*
	About dialog box dialog procedure
*/
INT_PTR APIENTRY AboutDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	BOOL bRet = FALSE;

	switch (message)
	{
		case WM_INITDIALOG:
			bRet = TRUE;
			break;

		case WM_COMMAND:
		{
			if (LOWORD(wParam) == IDOK ||
				LOWORD(wParam) == IDCANCEL)
			{
				EndDialog(hDlg, TRUE);
				bRet = TRUE;
			}
			break;
		}
	}

	return (INT_PTR)bRet;
}

/*
	Copy data from g_hText global variable to clipboard
*/
BOOL CopyClipboard(HWND hWnd)
{
	LPSTR lpszData;
	LPSTR lpszText;
	/* Allocate hData memory and copy g_hText into it */
	HANDLE hData = GlobalAlloc(GMEM_DDESHARE, GlobalSize(g_hText));
	if (hData == NULL) 
	{
		Error_OutOfMemory();
		return FALSE;
	}

	lpszData = (char*)GlobalLock(hData);
	if (lpszData == NULL) 
	{
		Error_OutOfMemory();
		return FALSE;
	}
	lpszText = GlobalLock(g_hText);
	if (lpszText == NULL) 
	{
		Error_OutOfMemory();
		return FALSE;
	}
	lstrcpyn(lpszData, lpszText, (int)GlobalSize(g_hText));
	GlobalUnlock(g_hText);
	GlobalUnlock(hData);

	/* Clear the clipboard and post hData. */
	if (OpenClipboard(hWnd)) 
	{
		EmptyClipboard();
#ifdef UNICODE
		SetClipboardData(CF_UNICODETEXT, hData);
#else
		SetClipboardData(CF_TEXT, hData);
#endif
		CloseClipboard();
	}
	return TRUE;
}

/*
	Paste data from clipboard into g_hText global variable
*/
BOOL PasteClipboard(HWND hWnd)
{
	if (OpenClipboard(hWnd)) 
	{
		HANDLE hData;
		LPSTR lpszData;
		LPSTR lpszText;
		HMENU hMenu;

		/* get text from the clipboard */
#ifdef UNICODE
		hData = GetClipboardData(CF_UNICODETEXT);
#else
		hData = GetClipboardData(CF_TEXT);
#endif
		if (hData == NULL) 
		{
			CloseClipboard();
			return FALSE;
		}
		if (g_hText != NULL) 
		{
			GlobalFree(g_hText);
		}
		g_hText = GlobalAlloc(GMEM_MOVEABLE, GlobalSize(hData));
		if (g_hText == NULL) 
		{
			Error_OutOfMemory();
			CloseClipboard();
			return FALSE;
		}

		lpszData = GlobalLock(hData);
		if (lpszData == NULL) 
		{
			Error_OutOfMemory();
			CloseClipboard();
			return FALSE;
		}
		lpszText = GlobalLock(g_hText);
		if (lpszText == NULL) 
		{
			Error_OutOfMemory();
			CloseClipboard();
			return FALSE;
		}
                        
		lstrcpyn(lpszText, lpszData, (int)GlobalSize(hData));
		GlobalUnlock(g_hText);
		GlobalUnlock(hData);

		CloseClipboard();

		/* Update menu */
		hMenu = GetMenu(hWnd);
		EnableMenuItem(hMenu, IDM_CUT, MF_ENABLED);
		EnableMenuItem(hMenu, IDM_COPY, MF_ENABLED);
		/* Update window */
		InvalidateRect(hWnd, NULL, TRUE);
		UpdateWindow(hWnd);
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

/*
	Menu actions
*/
BOOL ProcessAction(HWND hWnd, WORD id)
{
	switch (id) 
	{
		case IDM_ABOUT:
			DialogBox(g_hInstance, TEXT("AboutBox"), hWnd, AboutDlgProc);
			break;
		case IDM_EXIT:
			DestroyWindow(hWnd);
			break;
		/* edit menu commands */
		case IDM_CUT:
			if (g_hText != NULL)
			{
				HMENU hMenu;

				CopyClipboard(hWnd);
				GlobalFree(g_hText);
				g_hText = NULL;
				// Update menu
				hMenu = GetMenu(hWnd);
				EnableMenuItem(hMenu, IDM_CUT, MF_GRAYED);
				EnableMenuItem(hMenu, IDM_COPY, MF_GRAYED);
				// Update window
				InvalidateRect(hWnd, NULL, TRUE);
				UpdateWindow(hWnd);
			}
			break;
		case IDM_COPY:
			if (g_hText != NULL)
				CopyClipboard(hWnd);
			break;
		case IDM_PASTE:
			PasteClipboard(hWnd);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

/*
	Initialize menu
*/
BOOL UpdateMenuState(HWND hWnd, HMENU hMenu)
{
	if (hMenu == GetMenu(hWnd)) 
	{
		if (OpenClipboard(hWnd)) 
		{
#ifdef UNICODE
			if (IsClipboardFormatAvailable(CF_UNICODETEXT))
#else
			if (IsClipboardFormatAvailable(CF_TEXT) || IsClipboardFormatAvailable(CF_OEMTEXT))
#endif
			{
				EnableMenuItem(hMenu, IDM_PASTE, MF_ENABLED);
			}
			else
			{
				EnableMenuItem(hMenu, IDM_PASTE, MF_GRAYED);
			}
			CloseClipboard();
			return TRUE;
		}
		else /* Clipboard is not available */
		{
			return FALSE;
		}
	}
	return TRUE;
}

/*
	Paint the g_hText into the given HDC
*/
BOOL Paint(HWND hWnd, HDC hDC)
{
	if (g_hText != NULL) 
	{
		LPTSTR lpszText = GlobalLock(g_hText);
		if (lpszText == NULL) 
		{
			Error_OutOfMemory();
			return FALSE;
		} 
		else 
		{
			RECT rectClient;
			GetClientRect(hWnd, &rectClient);
			DrawText(hDC, lpszText, -1, &rectClient, 
						DT_EXTERNALLEADING | DT_NOPREFIX | DT_WORDBREAK);
						GlobalUnlock(g_hText);
		}
	}
	return TRUE;
}

/*
	Window procedure for the main window
*/
LRESULT APIENTRY MainWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) 
	{
		case WM_INITMENU:
			UpdateMenuState(hWnd, (HMENU)wParam);
			break;

		case WM_COMMAND:
			ProcessAction(hWnd, LOWORD(wParam));
			break;

		case WM_SIZE:
			InvalidateRect(hWnd, NULL, TRUE);
			break;

		case WM_PAINT:
		{
			PAINTSTRUCT ps = {0};
			HDC hDC = BeginPaint(hWnd, &ps);
			Paint(hWnd, hDC);
			EndPaint (hWnd, &ps);
			break;
		}
		case WM_DESTROY:
			PostQuitMessage(0);
			break;

		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
    }
    return 0;
}

/*
	Register Window Class
*/
BOOL InitApplication(HINSTANCE hInstance)
{
	BOOL success;
	/* Register "ClipboardWC" WNDCLASS */
	WNDCLASS  wc = {0};

	wc.lpfnWndProc = (WNDPROC) MainWndProc;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1); 
	wc.lpszMenuName =  TEXT("ClipboardMenu");
	wc.lpszClassName = TEXT("ClipboardWC");

	success = RegisterClass(&wc);
	/* Set global variable */
	g_hInstance = hInstance;
	/* Load accelerators */
	g_hAccTable = LoadAccelerators(g_hInstance, TEXT("ClipboardAcc"));

	return success;
}

BOOL InitText()
{
	BOOL success = TRUE;
	/* Copy s_szSampleText into g_hText */
	g_hText = GlobalAlloc(GMEM_MOVEABLE, sizeof(s_szSampleText));
	if (g_hText == NULL)
	{
		Error_OutOfMemory();
		success = FALSE;
	}
	else
	{
		LPSTR lpszText = GlobalLock(g_hText);
		if (lpszText == NULL) 
		{
			Error_OutOfMemory();
			success = FALSE;
		}
		else
		{	
			lstrcpyn(lpszText, (LPSTR)s_szSampleText, sizeof(s_szSampleText));
			GlobalUnlock(lpszText);
		}
	}
	return success;
}

/*
	Create main window
*/
BOOL CreateMainWindow(HINSTANCE hInstance, INT nCmdShow)
{
	BOOL success;

	g_hWnd = CreateWindow(
					TEXT("ClipboardWC"),
					TEXT("Clipboard Sample Application"),
					WS_OVERLAPPEDWINDOW,
					CW_USEDEFAULT, 0, CW_USEDEFAULT, 0,
					NULL, NULL, hInstance, NULL);
	if (g_hWnd == NULL)
	{
		success = FALSE;
	}
	else
	{
		ShowWindow(g_hWnd, nCmdShow);
		success = UpdateWindow(g_hWnd);
	}
    return success;
}

/*
	Program entry point
*/
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG msg = {0};
	BOOL success = TRUE;

	/* Window class and resources */
	if (!InitApplication(hInstance))
	{
		return 0;
	}
	/* Initialize g_hText */
	if (!InitText())
	{
		return 0;
	}
	/* Main window */
	if (!CreateMainWindow(hInstance, nCmdShow))
	{
		return 0;
	}
	/* Event loop */
	while (TRUE)
	{
		BOOL bRet = GetMessage(&msg, NULL, 0, 0 );
		if (bRet > 0)
		{
			/* Check for accelerator messages */
			if (!TranslateAccelerator(g_hWnd, g_hAccTable, &msg))
			{
				/* Not accelerator message. Translate and dispatch */
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		else
		{
			if (bRet < 0)
				success = FALSE;
			break;
		}
	}
	/* Exit */
	if (success)
	{
		return (int)msg.wParam;
	}
	else
	{
		return 0;
	}
}
